/*
 * opalvxml.h
 *
 * Header file for IVR code
 *
 * A H.323 IVR application.
 *
 * Copyright (C) 2002 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: opalvxml.h,v $
 * Revision 1.8  2002/07/29 15:13:51  craigs
 * Added autodelete option to PlayFile
 *
 * Revision 1.7  2002/07/18 04:17:12  robertj
 * Moved virtuals to source and changed name of G.723.1 file capability
 *
 * Revision 1.6  2002/07/10 13:16:19  craigs
 * Moved some VXML classes from Opal back into PTCLib
 * Added ability to repeat outputted data
 *
 * Revision 1.5  2002/07/09 08:27:23  craigs
 * Added GetMediaFormat to channels
 *
 * Revision 1.4  2002/07/03 01:53:58  craigs
 * Added newline to the end of the file for Linux
 *
 * Revision 1.3  2002/07/02 06:32:04  craigs
 * Added recording functions
 *
 * Revision 1.2  2002/06/28 02:42:54  craigs
 * Fixed problem with G.723.1 file codec not identified as native format
 *
 * Revision 1.1  2002/06/27 05:44:39  craigs
 * Initial version
 *
 * Revision 1.2  2002/06/26 09:05:28  csoutheren
 * Added ability to utter various "sayas" types within prompts
 *
 * Revision 1.1  2002/06/26 01:13:53  csoutheren
 * Disassociated VXML and Opal/OpenH323 specific elements
 *
 *
 */

#ifndef _OpenIVR_OPALVXML_H
#define _OpenIVR_OPALVXML_H

#include <ptclib/vxml.h>
#include <opalwavfile.h>
#include <ptclib/delaychan.h>
#include <h323caps.h>
#include <h245.h>
#include <h323con.h>

class G7231_File_Codec : public H323AudioCodec
{
  PCLASSINFO(G7231_File_Codec, H323AudioCodec);

  public:
    G7231_File_Codec(Direction dir);

    unsigned GetBandwidth() const;
    static int GetFrameLen(int val);
      
    BOOL Read(BYTE * buffer, unsigned & length, RTP_DataFrame &);
    BOOL Write(const BYTE * buffer, unsigned length, const RTP_DataFrame & rtp, unsigned & frames);

    BOOL IsRawDataChannelNative() const
      { return TRUE; }


  protected:
    PAdaptiveDelay incomingDelay;
};  

class G7231_File_Capability : public H323AudioCapability
{
  PCLASSINFO(G7231_File_Capability, H323AudioCapability)

  public:
    G7231_File_Capability();

    unsigned GetSubType() const;
    PString GetFormatName() const;

    H323Codec * CreateCodec(H323Codec::Direction direction) const;

    BOOL OnSendingPDU(H245_AudioCapability & cap, unsigned packetSize) const;
    BOOL OnReceivedPDU(const H245_AudioCapability & pdu, unsigned & packetSize);
    PObject * Clone() const;
};


//////////////////////////////////////////////////////////////////

class OpalVXMLSession;

class OpalVXMLOutgoingChannel : public PVXMLOutgoingChannel
{
  PCLASSINFO(OpalVXMLOutgoingChannel, PVXMLOutgoingChannel);

  public:
    OpalVXMLOutgoingChannel(OpalVXMLSession & vxml);

    // new functions
    PWAVFile * CreateWAVFile(const PFilePath & fn);
    virtual unsigned GetWavFileType() const = 0;
    virtual OpalMediaFormat GetMediaFormat() const = 0;

  protected:
    virtual BOOL IsWAVFileValid(OpalWAVFile & chan) = 0;
};

class OpalVXMLOutgoingChannelPCM : public OpalVXMLOutgoingChannel
{
  PCLASSINFO(OpalVXMLOutgoingChannelPCM, OpalVXMLOutgoingChannel);

  public:
    OpalVXMLOutgoingChannelPCM(OpalVXMLSession & vxml);

    unsigned GetWavFileType() const
     { return PWAVFile::PCM_WavFile; }

    OpalMediaFormat GetMediaFormat() const
      { return OPAL_PCM16; }

  protected:
    // overrides from PVXMLOutgoingChannel
    void DelayFrame(PINDEX len);
    BOOL IsWAVFileValid(OpalWAVFile & chan);
    BOOL ReadFrame(PINDEX amount);
    void CreateSilenceFrame(PINDEX amount);
};

class OpalVXMLOutgoingChannelG7231 : public OpalVXMLOutgoingChannel
{
  PCLASSINFO(OpalVXMLOutgoingChannelG7231, OpalVXMLOutgoingChannel);
  public:
    OpalVXMLOutgoingChannelG7231(OpalVXMLSession & vxml);
    void QueueFile(const PString & ofn, PINDEX repeat = 1, PINDEX delay = 0);

    OpalMediaFormat GetMediaFormat() const
      { return OPAL_G7231; }

    unsigned GetWavFileType() const
      { return PWAVFile::fmt_VivoG7231; }

  protected:
    // overrides from PVXMLOutgoingChannel
    void DelayFrame(PINDEX len);
    BOOL IsWAVFileValid(OpalWAVFile & chan);
    BOOL ReadFrame(PINDEX amount);
    void CreateSilenceFrame(PINDEX amount);
};

//////////////////////////////////////////////////////////////////

class OpalVXMLIncomingChannel : public PVXMLIncomingChannel
{
  PCLASSINFO(OpalVXMLIncomingChannel, PVXMLIncomingChannel)

  public:
    OpalVXMLIncomingChannel(OpalVXMLSession & vxml);

    PWAVFile * CreateWAVFile(const PFilePath & fn);
    virtual unsigned GetWavFileType() const = 0;
    virtual OpalMediaFormat GetMediaFormat() const = 0;
};

class OpalVXMLIncomingChannelPCM : public OpalVXMLIncomingChannel
{
  PCLASSINFO(OpalVXMLIncomingChannelPCM, OpalVXMLIncomingChannel)

  public:
    OpalVXMLIncomingChannelPCM(OpalVXMLSession & vxml);

    unsigned GetWavFileType() const
      { return PWAVFile::PCM_WavFile; }

    OpalMediaFormat GetMediaFormat() const
      { return OPAL_PCM16; }

  protected:
    // overrides from PVXMLIncomingChannel
    void DelayFrame(PINDEX len);
    BOOL WriteFrame(const void * buf, PINDEX len);
};

class OpalVXMLIncomingChannelG7231 : public OpalVXMLIncomingChannel
{
  PCLASSINFO(OpalVXMLIncomingChannelG7231, PVXMLIncomingChannel);

  public:
    OpalVXMLIncomingChannelG7231(OpalVXMLSession & vxml);

    unsigned GetWavFileType() const
      { return PWAVFile::fmt_MSG7231; }

    OpalMediaFormat GetMediaFormat() const
      { return OPAL_G7231; }

  protected:
    // overrides from PVXMLIncomingChannel
    void DelayFrame(PINDEX len);
    BOOL WriteFrame(const void * buf, PINDEX len);
};

//////////////////////////////////////////////////////////////////

class OpalVXMLSession : public PVXMLSession 
{
  PCLASSINFO(OpalVXMLSession, PVXMLSession);
  public:
    OpalVXMLSession();

    // overrides from PVXMLSession
    BOOL PlayFile(const PString & fn, PINDEX repeat = 1, PINDEX delay = 0, BOOL autoDelete = FALSE);
    BOOL PlayData(const PBYTEArray & data, PINDEX repeat = 1, PINDEX delay = 0);
    BOOL IsPlaying() const;

    BOOL StartRecording(const PFilePath & fn);
    BOOL EndRecording();
    BOOL IsRecording() const;

    // new functions
    virtual PVXMLOutgoingChannel * CreateOutgoingAudioChannel(H323AudioCodec & codec);
    virtual PVXMLIncomingChannel * CreateIncomingAudioChannel(H323AudioCodec & codec);

    virtual PVXMLIncomingChannel * SetIncomingAudioChannel(PVXMLIncomingChannel * chan);
    virtual PVXMLOutgoingChannel * SetOutgoingAudioChannel(PVXMLOutgoingChannel * chan);

  protected:
    PVXMLIncomingChannel * incomingChannel;
    PVXMLOutgoingChannel * outgoingChannel;
};

#endif

