/*
 * vxml.cxx
 *
 * VXML control for for Opal
 *
 * A H.323 IVR application.
 *
 * Copyright (C) 2002 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: opalvxml.cxx,v $
 * Revision 1.11  2002/07/29 15:10:36  craigs
 * Added autodelete option to PlayFile
 *
 * Revision 1.10  2002/07/29 12:54:42  craigs
 * Removed usages of cerr
 *
 * Revision 1.9  2002/07/18 04:17:12  robertj
 * Moved virtuals to source and changed name of G.723.1 file capability
 *
 * Revision 1.8  2002/07/10 13:16:58  craigs
 * Moved some VXML classes from Opal back into PTCLib
 * Added ability to repeat outputted data
 *
 * Revision 1.7  2002/07/09 08:48:41  craigs
 * Fixed trace messages
 *
 * Revision 1.6  2002/07/05 06:34:04  craigs
 * Changed comments and trace messages
 *
 * Revision 1.5  2002/07/03 04:58:13  robertj
 * Changed for compatibility with older GNU compilers
 *
 * Revision 1.4  2002/07/02 06:32:51  craigs
 * Added recording functions
 *
 * Revision 1.3  2002/06/28 02:42:11  craigs
 * Fixed problem with G.723.1 file naming conventions
 * Fixed problem with incorrect file open mode
 *
 * Revision 1.2  2002/06/28 01:24:03  robertj
 * Fixe dproblem with compiling without expat library.
 *
 * Revision 1.1  2002/06/27 05:44:11  craigs
 * Initial version
 *
 * Revision 1.2  2002/06/26 09:05:28  csoutheren
 * Added ability to utter various "sayas" types within prompts
 *
 * Revision 1.1  2002/06/26 01:13:53  csoutheren
 * Disassociated VXML and Opal/OpenH323 specific elements
 *
 * Revision 1.2  2002/06/21 08:18:22  csoutheren
 * Added start of grammar handling
 *
 * Revision 1.1  2002/06/20 06:35:44  csoutheren
 * Initial version
 *
 */

#include <ptlib.h>

#if P_EXPAT

#include <lpc10codec.h>
#include <mscodecs.h>
#include <gsmcodec.h>
#include <ptclib/delaychan.h>
#include <ptclib/pwavfile.h>
#include <ptclib/memfile.h>

#include "opalvxml.h"

#define	G7231_SAMPLES_PER_BLOCK	240
#define	G7231_BANDWIDTH		      (6300/100)


///////////////////////////////////////////////////////////////

OpalVXMLSession::OpalVXMLSession()
{
  incomingChannel = NULL;
  outgoingChannel = NULL;
}

PVXMLIncomingChannel * OpalVXMLSession::CreateIncomingAudioChannel(H323AudioCodec & codec)
{
  BOOL mediaIsPCM = FALSE;
  OpalMediaFormat mediaFormat = codec.GetMediaFormat();

  if (!codec.IsRawDataChannelNative())
    mediaIsPCM = TRUE;
  else if (mediaFormat == OPAL_G7231_6k3)
    mediaIsPCM = FALSE;
  else {
    PTRACE(2, "OpalVXML\tUnknown media format " << mediaFormat);
    return NULL;
  }

  if (mediaIsPCM)
    return SetIncomingAudioChannel(new OpalVXMLIncomingChannelPCM(*this));

  return SetIncomingAudioChannel(new OpalVXMLIncomingChannelG7231(*this));
}

PVXMLOutgoingChannel * OpalVXMLSession::CreateOutgoingAudioChannel(H323AudioCodec & codec)
{
  BOOL mediaIsPCM = FALSE;
  OpalMediaFormat mediaFormat = codec.GetMediaFormat();

  if (!codec.IsRawDataChannelNative())
    mediaIsPCM = TRUE;
  else if (mediaFormat == OPAL_G7231_6k3)
    mediaIsPCM = FALSE;
  else {
    PTRACE(2, "OpalVXML\tUnknown media format " << mediaFormat);
    return NULL;
  }

  if (mediaIsPCM)
    return SetOutgoingAudioChannel(new OpalVXMLOutgoingChannelPCM(*this));

  return SetOutgoingAudioChannel(new OpalVXMLOutgoingChannelG7231(*this));
}

PVXMLIncomingChannel * OpalVXMLSession::SetIncomingAudioChannel(PVXMLIncomingChannel * chan)
{
  if (incomingChannel != NULL) {
    delete incomingChannel;
    incomingChannel = NULL;
  }

  incomingChannel = chan;
  return incomingChannel;
}

PVXMLOutgoingChannel * OpalVXMLSession::SetOutgoingAudioChannel(PVXMLOutgoingChannel * chan)
{
  if (outgoingChannel != NULL) {
    delete outgoingChannel;
    outgoingChannel = NULL;
  }

  outgoingChannel = chan;
  return outgoingChannel;
}

BOOL OpalVXMLSession::PlayFile(const PString & fn, PINDEX repeat, PINDEX delay, BOOL autoDelete)
{
  if (outgoingChannel != NULL)
    outgoingChannel->QueueFile(fn, repeat, delay, autoDelete);

  return TRUE;
}

BOOL OpalVXMLSession::PlayData(const PBYTEArray & data, PINDEX repeat, PINDEX delay)
{
  if (outgoingChannel != NULL)
    outgoingChannel->QueueData(data, repeat, delay);

  return TRUE;
}

BOOL OpalVXMLSession::IsPlaying() const
{
  return (outgoingChannel != NULL) && outgoingChannel->IsPlaying();
}

BOOL OpalVXMLSession::StartRecording(const PFilePath & fn)
{
  if (incomingChannel != NULL)
    return incomingChannel->StartRecording(fn);

  return FALSE;
}

BOOL OpalVXMLSession::EndRecording()
{
  if (incomingChannel != NULL)
    return incomingChannel->EndRecording();

  return FALSE;
}

BOOL OpalVXMLSession::IsRecording() const
{
  return (incomingChannel != NULL) && incomingChannel->IsRecording();
}

///////////////////////////////////////////////////////////////

BOOL CheckWAVFileValid(OpalWAVFile & chan, BOOL mustBePCM)
{
  // Check the wave file header
  if (!chan.IsValid()) {
    PTRACE(1, chan.GetName() << " wav file header invalid");
    return FALSE;
  }

  // Check the wave file format
  if (mustBePCM) {
    if (chan.GetFormat() != PWAVFile::fmt_PCM) {
      PTRACE(1, chan.GetName() << " is not a PCM format wav file");
      PTRACE(1, "It is format " << chan.GetFormat() );
      return FALSE;
    }

    if ((chan.GetSampleRate() != 8000) &&
        (chan.GetChannels() != 1) &&
        (chan.GetSampleSize() != 16)) {
      PTRACE(1, chan.GetName() << " is not a 16 Bit, Mono, 8000 Hz (8Khz) PCM wav file");
      PTRACE(1, "It is " << chan.GetSampleSize() << " bits, "
                         << (chan.GetChannels()==1 ? "mono " : "stereo ")
                         << chan.GetSampleRate() << " Hz");
      return FALSE;
    }
  }

  else if ((chan.GetFormat() != PWAVFile::fmt_MSG7231) && 
      (chan.GetFormat() != PWAVFile::fmt_VivoG7231)) {
    PTRACE(1, chan.GetName() << " is not a G.723.1 format wav file");
    PTRACE(1, "It is format " << chan.GetFormat() );
    return FALSE;
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////

OpalVXMLIncomingChannel::OpalVXMLIncomingChannel(OpalVXMLSession & vxml)
  : PVXMLIncomingChannel(vxml)
{
}

PWAVFile * OpalVXMLIncomingChannel::CreateWAVFile(const PFilePath & fn)
{ 
  return new OpalWAVFile(fn, PFile::WriteOnly, PFile::ModeDefault, GetWavFileType()); 
}

///////////////////////////////////////////////////////////////

OpalVXMLIncomingChannelPCM::OpalVXMLIncomingChannelPCM(OpalVXMLSession & vxml)
  : OpalVXMLIncomingChannel(vxml)
{
}

BOOL OpalVXMLIncomingChannelPCM::WriteFrame(const void * buf, PINDEX len)
{
  //cerr << "Writing PCM " << len << endl;
  return wavFile->Write(buf, len);
}

void OpalVXMLIncomingChannelPCM::DelayFrame(PINDEX len)
{
  delay.Delay(len/16);
}

///////////////////////////////////////////////////////////////
// Override some of the IncomingChannelPCM functions to write
// G723.1 data instead of PCM data.

OpalVXMLIncomingChannelG7231::OpalVXMLIncomingChannelG7231(OpalVXMLSession & vxml)
  : OpalVXMLIncomingChannel(vxml)
{
}

BOOL OpalVXMLIncomingChannelG7231::WriteFrame(const void * buf, PINDEX /*len*/)
{
  int frameLen = G7231_File_Codec::GetFrameLen(*(BYTE *)buf);
  return wavFile->Write(buf, frameLen);
}

void OpalVXMLIncomingChannelG7231::DelayFrame(PINDEX /*len*/)
{
  // Ignore the len parameter as that is the compressed size.
  // We must delay by the actual sample time.
  delay.Delay((G7231_SAMPLES_PER_BLOCK*2)/16);
}

///////////////////////////////////////////////////////////////

OpalVXMLOutgoingChannel::OpalVXMLOutgoingChannel(OpalVXMLSession & _vxml)
  : PVXMLOutgoingChannel(_vxml)
{
}

PWAVFile * OpalVXMLOutgoingChannel::CreateWAVFile(const PFilePath & fn)
{ 
  OpalWAVFile * file = new OpalWAVFile(fn, PFile::ReadOnly, PFile::ModeDefault, GetWavFileType()); 
  if (!IsWAVFileValid(*file)) {
    delete file;
    file = NULL;
  }

  return file;
}

///////////////////////////////////////////////////////////////

OpalVXMLOutgoingChannelPCM::OpalVXMLOutgoingChannelPCM(OpalVXMLSession & vxml)
  : OpalVXMLOutgoingChannel(vxml)
{
}

BOOL OpalVXMLOutgoingChannelPCM::IsWAVFileValid(OpalWAVFile & chan) 
{
  return CheckWAVFileValid(chan, TRUE);
}

void OpalVXMLOutgoingChannelPCM::DelayFrame(PINDEX amount)
{
  delay.Delay(amount / 16);
}

BOOL OpalVXMLOutgoingChannelPCM::ReadFrame(PINDEX amount)
{
  frameOffs = 0;
  frameLen  = amount;

  BOOL result = PIndirectChannel::Read(frameBuffer.GetPointer(), frameLen);

  // if we did not read a full frame of audio, fill the end of the
  // frame with zeros.
  PINDEX count = GetLastReadCount();
  if (count < frameLen)
    memset(frameBuffer.GetPointer()+count, 0, frameLen-count);

  return result;
}

void OpalVXMLOutgoingChannelPCM::CreateSilenceFrame(PINDEX amount)
{
  frameOffs = 0;
  frameLen  = amount;
  memset(frameBuffer.GetPointer(), 0, frameLen);
}

///////////////////////////////////////////////////////////////

OpalVXMLOutgoingChannelG7231::OpalVXMLOutgoingChannelG7231(OpalVXMLSession & vxml)
  : OpalVXMLOutgoingChannel(vxml)
{
}

void OpalVXMLOutgoingChannelG7231::QueueFile(const PString & ofn, PINDEX repeat, PINDEX delay)
{
  PString fn = ofn;;

  // add in _g7231 prefix, if not there already
  PINDEX pos = ofn.FindLast('.');
  if (pos == P_MAX_INDEX) {
    if (fn.Right(6) != "_g7231")
      fn += "_g7231";
  } else {
    PString basename = ofn.Left(pos);
    PString ext      = ofn.Mid(pos+1);
    if (basename.Right(6) != "_g7231")
      basename += "_g7231";
    fn = basename + "." + ext;
  }
  PVXMLOutgoingChannel::QueueFile(fn, repeat, delay);
}

BOOL OpalVXMLOutgoingChannelG7231::IsWAVFileValid(OpalWAVFile & chan) 
{
  return CheckWAVFileValid(chan, FALSE);
}

void OpalVXMLOutgoingChannelG7231::DelayFrame(PINDEX /*amount*/)
{
  delay.Delay(30);
}

BOOL OpalVXMLOutgoingChannelG7231::ReadFrame(PINDEX /*amount*/)
{
  if (!PIndirectChannel::Read(frameBuffer.GetPointer(), 1))
    return FALSE;

  frameOffs = 0;
  frameLen = G7231_File_Codec::GetFrameLen(frameBuffer[0]);

  return PIndirectChannel::Read(frameBuffer.GetPointer()+1, frameLen-1);
}

void OpalVXMLOutgoingChannelG7231::CreateSilenceFrame(PINDEX /*amount*/)
{
  frameOffs = 0;
  frameLen  = 4;

  frameBuffer[0] = 2;
  memset(frameBuffer.GetPointer()+1, 0, 3);
}

///////////////////////////////////////////////////////////////
//  : H323AudioCapability(12*G7231_BLOCK_SIZE, 4*G7231_BLOCK_SIZE)

G7231_File_Capability::G7231_File_Capability()
  : H323AudioCapability(8, 4)
{
}

unsigned G7231_File_Capability::GetSubType() const
{
  return H245_AudioCapability::e_g7231;
}

PString G7231_File_Capability::GetFormatName() const
{
  return "G.723.1{file}";
}

BOOL G7231_File_Capability::OnSendingPDU(H245_AudioCapability & cap, unsigned packetSize) const
{
  // set the choice to the correct type
  cap.SetTag(GetSubType());

  // get choice data
  H245_AudioCapability_g7231 & g7231 = cap;

  // max number of audio frames per PDU we want to send
  g7231.m_maxAl_sduAudioFrames = packetSize; 

  // no silence suppression
  g7231.m_silenceSuppression = FALSE;

  return TRUE;
}

BOOL G7231_File_Capability::OnReceivedPDU(const H245_AudioCapability & cap, unsigned & packetSize)
{
  const H245_AudioCapability_g7231 & g7231 = cap;
  packetSize = g7231.m_maxAl_sduAudioFrames;
  return TRUE;
}

PObject * G7231_File_Capability::Clone() const
{
  return new G7231_File_Capability(*this);
}

H323Codec * G7231_File_Capability::CreateCodec(H323Codec::Direction direction) const
{
  return new G7231_File_Codec(direction);
}

///////////////////////////////////////////////////////////////

G7231_File_Codec::G7231_File_Codec(Direction dir)
  : H323AudioCodec(OPAL_G7231_6k3, dir)
{
}


int G7231_File_Codec::GetFrameLen(int val)
{
  static const int frameLen[] = { 24, 20, 4, 1 };
  return frameLen[val & 3];
}

BOOL G7231_File_Codec::Read(BYTE * buffer, unsigned & length, RTP_DataFrame &)
{
  if (rawDataChannel == NULL)
    return FALSE;
    
  if (!rawDataChannel->Read(buffer, 1)) {
    PTRACE(2, "G7231_File_Codec\tfirst read failed");
    return FALSE;
  }
  if (rawDataChannel->GetLastReadCount() == 0) {
    PTRACE(3, "G7231_File_Codec\tfirst read count is zero");
    return FALSE;
  }

  int readLen = G7231_File_Codec::GetFrameLen(buffer[0]);

  if (readLen > 0) {
    if (!rawDataChannel->Read(buffer+1, readLen-1)) {
      PTRACE(2, "G7231_File_Codec\tsecond read failed");
      return FALSE;
    }
    if (rawDataChannel->GetLastReadCount() == 0) {
      PTRACE(3, "G7231_File_Codec\tsecond read count is zero");
      return FALSE;
    }
  }

  length = readLen;
  return TRUE;
}

BOOL G7231_File_Codec::Write(const BYTE * buffer, unsigned length, const RTP_DataFrame & /* rtp */, unsigned & writtenLength)
{
  if (rawDataChannel == NULL)
    return TRUE;

  static const BYTE silence[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                  0, 0, 0, 0};

  // If the length is zero, output silence to the file..
  if (length == 0) {
    PTRACE(5,"G.723.1 zero length frame");
    writtenLength = 0;
    return rawDataChannel->Write(silence, 24);
  }

  int writeLen;
  switch (buffer[0]&3) {
    case 0:
      writeLen = 24;
      break;
    case 1:
      writeLen = 20;
      break;
    case 2:
      // Windows Media Player cannot play 4 byte SID (silence) frames.
      // So write out a 24 byte frame of silence instead.
      PTRACE(1,"Replacing SID with 24 byte frame");
      writtenLength = 4;
      return rawDataChannel->Write(silence, 24);
//    writeLen = 4;
//    break;
    default :
      writeLen = 1;
      break;
  }

  PTRACE(4,"G.723.1 frame length = " <<writeLen);

  writtenLength = writeLen;

  return rawDataChannel->Write(buffer, writeLen);
}

unsigned G7231_File_Codec::GetBandwidth() const
{ 
  return G7231_BANDWIDTH; 
}


///////////////////////////////////////////////////////////////


#endif
